* chantillonnage - Confrence - 2022_11_V3.sas ;
* 2022-10-21 : Version AVEC SURVEYSELECT et utilisant la table BASEBALL;
* 2022-10-27 : Tests de performance du SRS pour un DATA STEP VIEW;
* 2022-11-06 : Ajout exemple de Bootstrap en annexe;
* 2022-11-08 : Modifi chantillonnage stratifi via DATA
               (mthode 1) car PROC FREQ est plus simple
                comprendre, malgr une tape de plus.  Supprim aussi
               la variable macro TOT_SAMPLE, car ajout une "complexit"
               inutile;



* 1) Simple Random Sampling without replacement, 
     exact sample size;
* 1_0) Method very simple but not very useful since sample
       is of approximate size;
DATA SIMP_0;
  SET SASHELP.BASEBALL(KEEP=NAME  TEAM  LEAGUE
                            NHITS  NATBAT)
      NOBS=_N;
  IF RAND("UNIFORM") <= 50 / _N;
RUN;

* 1_1) Using DATA step (cumbersome and inefficient);
DATA TEMP;
  SET SASHELP.BASEBALL(KEEP=NAME  TEAM  LEAGUE  NHITS  NATBAT);
  RD = RAND("UNIFORM");
RUN;
PROC SORT  DATA=TEMP  OUT=TEMPA; /* OBS= pas utilisable en sortie */
  BY RD;
RUN;
DATA SIMP_1;
  SET TEMPA(OBS=50 DROP=RD);
RUN;

* 1_2) Using DATA step (efficient but more complex);
DATA SIMP_2(DROP=_K);
   SET SASHELP.BASEBALL(KEEP=NAME  TEAM  LEAGUE  NHITS  NATBAT)
       NOBS=_N;
   RETAIN _K  50  SelectionProb  SamplingWeight;
   IF _N_=1 THEN DO;                      /* Optional */
     SelectionProb = 50 / _N;    /* Optional */
     SamplingWeight = _N / 50;   /* Optional */
   END;                                   /* Optional */
   IF RAND ("UNIFORM") < _K / _N THEN DO;
      OUTPUT;
      _K = _K - 1;
   END;
   _N = _N - 1;
   IF _K = 0 THEN STOP; /* Add-on to make more efficient */
RUN;

* 1_2A) Same as before, but showing values of _K and _N
        (for both selected and unselected values - optional
        section skipped here);
OPTIONS PAGESIZE=200;
DATA SIMP_2(DROP=_K);
   SET SASHELP.BASEBALL(KEEP=NAME  TEAM  LEAGUE  NHITS  NATBAT)
       NOBS=_N;
   RETAIN _K  50;
   PUT @1 NAME=  @26 TEAM=  @45 _K=  @52 _N=;
   IF RAND ("UNIFORM") < _K / _N THEN DO;
      OUTPUT;
      _K = _K - 1;
   END;
   _N = _N - 1;
   IF _K = 0 THEN STOP; /* Add-on to make it more efficient */
RUN;

* 1_2B) Using binomial probability distribution rather than
        uniform;
DATA SIMP_2(DROP=_K  PROB);
   SET SASHELP.BASEBALL(KEEP=NAME  TEAM  LEAGUE  NHITS  NATBAT)
       NOBS=_N;
   RETAIN _K  50;
   PROB = _K / _N;
   IF RAND ("BINOMIAL",PROB,1) THEN DO;
      OUTPUT;
      _K = _K - 1;
   END;
   _N = _N - 1;
   IF _K = 0 THEN STOP; /* Add-on to make it more efficient */
RUN;

* 1_3) Using SQL;
PROC SQL  OUTOBS=50;
  CREATE TABLE SIMP_3 AS
    SELECT NAME, TEAM, LEAGUE, NHITS, NATBAT
    FROM SASHELP.BASEBALL
    ORDER BY RAND("UNIFORM");
QUIT;

* 1_4) Using PROC SURVEYSELECT;
PROC SURVEYSELECT  DATA=SASHELP.BASEBALL(KEEP=NAME  TEAM  LEAGUE  NHITS  NATBAT)
                   SAMPSIZE=50 /* N= est un synonyme */
				   METHOD=SRS
                   OUT=SIMP_4;
RUN;


* 2) SRS STRATIFIED (ONE VAR.) - EXACT SAMPLE SIZE;
* 2_1) Using DATA step;
PROC SORT  DATA=SASHELP.BASEBALL(KEEP=NAME  TEAM  LEAGUE  NHITS  NATBAT)
           OUT=BASEBALL;
  BY LEAGUE;
RUN;
* The combination of PROC FREQ + DATA is not much long
  and easier to understand than a single DATA step or a
  PROC SUMMARY + DATA combination to achieve same results;
PROC FREQ  DATA=BASEBALL  NOPRINT;
  TABLE LEAGUE / OUT=_STATS(RENAME=(COUNT=STRA_SIZE));
RUN;
DATA _STATS(DROP=PERCENT);
  SET _STATS;
  STRA_SAMPLE = ROUND((PERCENT/100)*50);
RUN;
DATA STRA_1(DROP=_K  _N  STRA_SAMPLE  STRA_SIZE);
  MERGE BASEBALL
        _STATS;
  BY LEAGUE;
  RETAIN _K _N;
  IF FIRST.LEAGUE THEN DO;
     _K = STRA_SAMPLE;
	 _N = STRA_SIZE;
  END;
  IF RAND("UNIFORM") < _K / _N THEN DO;
     OUTPUT;
     _K = _K - 1;
  END;
  _N = _N - 1;
RUN;

* 2_2) Using PROC SURVEYSELECT - la table en entre DOIT tre trie;
* 2_2A) Ajouter ALLOC=PROP dans STRATA asssure que la
        taille totale d'chantillon demande (ici 50)
        est divise en proportion de la taille des strates
       (soit 27 pour la ligue Amric. et 23 pour la Nation.);
PROC SURVEYSELECT  DATA=BASEBALL
                   SAMPSIZE=50 /* N= est un synonyme */
				   METHOD=SRS
                   OUT=STRA_2;
  STRATA LEAGUE / ALLOC=PROP;
RUN;

* 2_2B) Autre mthode: fournir la taille de chaque strate 
        comme une srie de valeurs entre parenthses, dans
        l'ordre croissant des valeurs formates de la 
        variable de stratification;
PROC SURVEYSELECT  DATA=BASEBALL
                   SAMPSIZE=(27 23)
				   METHOD=SRS
                   OUT=STRA_2;
  STRATA LEAGUE;
RUN;

* 2_2C) Autre mthode: fournir la taille de chaque strate
        via une table secondaire, trie selon les niveaux de
        la variable de stratification, et contenant la variable
        de stratification ainsi qu'une variable nomme obli-
        gatoirement "SampleSize" ou "_NSIZE_";
DATA T_REQUIRED;
  INPUT LEAGUE $8. +1 SampleSize 2.; /* Variable can also be named _NSIZE_ */
DATALINES;
American 27
National 23
RUN;  
PROC SURVEYSELECT  DATA=BASEBALL
                   SAMPSIZE=T_REQUIRED
				   METHOD=SRS
                   OUT=STRA_2;
  STRATA LEAGUE;
RUN;



* 3) RANDOMLY SPLIT A DATASET INTO "TRAINING" VS "VALIDATION" GROUPS;
* 3_1) Using DATA step:
          > we can split into more than two "groups",
          > we can store each "group" (calibration vs validation) into separate
            datasets, 
          > possibly other benefits also... 
          > size of each "group" is approximate;
DATA SPLIT_1;
  SET SASHELP.BASEBALL(KEEP=NAME  TEAM  LEAGUE  NHITS  NATBAT);
  SELECTED = 1;
  IF RAND("UNIFORM") <= .4 THEN SELECTED=0;
RUN;
* Following step is for verification purpose only;
PROC FREQ  DATA=SPLIT_1;
  TABLE SELECTED / NOCUM;
RUN; 
 
* 3_2) Using PROC SQL;
PROC SQL;
  CREATE TABLE SPLIT_2 AS
  SELECT NAME, TEAM, LEAGUE, NHITS, NATBAT,
         CASE 
            WHEN (RAND("UNIFORM") <= .4) THEN 0
			ELSE                              1
	     END AS SELECTED
  FROM SASHELP.BASEBALL;
RUN;
* Following step is for verification purpose only;
PROC FREQ  DATA=SPLIT_2;
  TABLE SELECTED / NOCUM;
RUN; 
 
* 3_3) Using PROC SURVEYSELECT;
* 3_3A) Two "groups":
           > both "groups" are part of the output dataset but an extra data step
             will separate them
           > size of each "group" is exact;
PROC SURVEYSELECT  DATA=SASHELP.BASEBALL(KEEP=NAME  TEAM  LEAGUE  NHITS  NATBAT)
                   SAMPRATE=.6 
                   METHOD=SRS
                   OUTALL
                   OUT=SPLIT_3;
RUN;
* Following step is for verification purpose only;
PROC FREQ  DATA=SPLIT_3;
  TABLE TEAM*SELECTED / NOFREQ NOCOL NOPERCENT;
RUN;  

* 3_3B) Three "groups" or more:
           > all "groups" are part of the output dataset
           > a variable named "GROUP" is added to output dataseet
           > size of each "group" is exact;
PROC SURVEYSELECT  DATA=SASHELP.BASEBALL(KEEP=NAME  TEAM  LEAGUE  NHITS  NATBAT)
                   GROUPS=(194 64 64)
                   OUT=SPLIT_3B;
RUN;
* Following step is for verification purpose only;
PROC FREQ  DATA=SPLIT_3B;
  TABLE TEAM*GroupID / NOFREQ NOCOL NOPERCENT;
RUN;  

* 3_3C) Three "groups" or same size:
           > all "groups" are part of the output dataset
           > a variable named "GROUP" is added to output dataseet;
PROC SURVEYSELECT  DATA=SASHELP.BASEBALL(KEEP=NAME  TEAM  LEAGUE  NHITS  NATBAT)
                   GROUPS=3
                   OUT=SPLIT_3C;
RUN;
* Following step is for verification purpose only;
PROC FREQ  DATA=SPLIT_3C;
  TABLE TEAM*GroupID / NOFREQ NOCOL NOPERCENT;
RUN;  

* 3_3D) Randomly split the "teams", not the individual players, into two
        "groups", so that each player of a team goes with its team - not
        shown in presentation (variables cannot be dropped in this case,
        which seems very strange to me);
PROC SURVEYSELECT  DATA=SASHELP.BASEBALL
                   SAMPRATE=.6 
                   METHOD=SRS
                   OUTALL
                   OUT=SPLIT_3D;
   SAMPLINGUNIT TEAM; 
RUN;
* Following step is for verification purpose only;
PROC FREQ  DATA=SPLIT_3D;
  TABLE TEAM*SELECTED / NOFREQ NOCOL NOPERCENT;
RUN; 



* 5) BOOTSTRAP;
* 5.1) Bootstrap using PROC SURVEYSELECT;
PROC SURVEYSELECT  DATA=SASHELP.BASEBALL(KEEP=NAME  TEAM  LEAGUE  NHITS  NATBAT)
                   SAMPSIZE=50
				   METHOD=SRS
                   OUT=BOOT_1
                   REPS=1000;
RUN;

* 5.2) Bootstrap using DATA steps (same algorithm as 1_2 above
       and checking if each player have been selected about the
       same number of times);
%MACRO BOOTSTRAP(NS=);
OPTIONS NONOTES;
%DO I=1 %TO &NS;
DATA TEMP(DROP=_K);
   SET SASHELP.BASEBALL(KEEP=NAME  TEAM  LEAGUE  NHITS  NATBAT)
       NOBS=_N;
   PLAYER_RANK = _N_;
   RETAIN _K  50  Replicate &I;
   IF RAND ("UNIFORM") < _K / _N THEN DO;
      OUTPUT;
      _K = _K - 1;
   END;
   _N = _N - 1;
RUN;
PROC APPEND  BASE=BOOT_2  DATA=TEMP;
RUN;
%END;
OPTIONS NOTES;
%MEND BOOTSTRAP;
%BOOTSTRAP(NS=1000);
* Frquence de slection de chaque joueur selon
  A) son nom et 
  B) son rang dans la table en entre (nomme PLAYER_RANK);
PROC FREQ  DATA=BOOT_2;
  TABLE NAME PLAYER_RANK / NOCUM;
RUN;



*===================================================================;
* 6) EFFICIENCY COMPARAISON FOR SIMPLE RANDOM SAMPLING,
     BETWEEN VARIOUS METHODS RELYING ON DATA STEP OR SQL;

* Assign library with a table of 5 000 000 observations
  and 50 variables (TEST_SORT_MASTER);
LIBNAME BIG_DATA "C:\App\Consult_client\TESTS_PERFORMANCE\Data";
%LET TOT_SAMPLE = 1000;

* 6A) Using DATA step (cumbersome and inefficient) - 9.5 secondes CPU;
DATA TEMP;
  SET BIG_DATA.TEST_SORT_MASTER;
  RD = RAND("UNIFORM");
RUN;
PROC SORT  DATA=TEMP  OUT=TEMPA(DROP=RD);   
   BY RD;
RUN;
DATA RAND6A_DATA;
  SET TEMPA(OBS=&TOT_SAMPLE);
RUN;

* 6B) Using DATA step view (better ??) - 15 secondes;
DATA TEMP2 / VIEW=TEMP2;
  SET BIG_DATA.TEST_SORT_MASTER;
  RD = RAND("UNIFORM");
RUN;
PROC SORT  DATA=TEMP2  OUT=TEMP2A(DROP=RD); 
  BY RD;
RUN;
DATA RAND6B_DATA;
  SET TEMP2A(OBS=&TOT_SAMPLE);
RUN;

* 6C) Using DATA step (efficient but more complex) - 0.4 secondes;
DATA RAND6C_DATA(DROP=_K);
   SET BIG_DATA.TEST_SORT_MASTER
       NOBS=_N;
   RETAIN _K  &TOT_SAMPLE  SelectionProb  SamplingWeight;
   IF _N_=1 THEN DO;                      /* Optional */
     SelectionProb = &TOT_SAMPLE / _N;    /* Optional */
     SamplingWeight = _N / &TOT_SAMPLE;   /* Optional */
   END;                                   /* Optional */
   IF RAND ("UNIFORM") < _K / _N THEN DO;
      OUTPUT;
      _K = _K - 1;
   END;
   _N = _N - 1;
   IF _K = 0 THEN STOP; /* Add-on to make more efficient */
RUN;

* 6D) Using PROC SQL - 8 seconds;
PROC SQL  OUTOBS=&TOT_SAMPLE;
  CREATE TABLE RAND6D_SQL AS
    SELECT *
    FROM BIG_DATA.TEST_SORT_MASTER
    ORDER BY RAND("UNIFORM");
QUIT;



